from logging import warning
from pathlib import Path
from typing import Optional
from warnings import warn

import numpy as np
from extra_data import RunDirectory, SourceNameError


def get_mem_cell_pattern(run, sources) -> np.ndarray:
    """Load the memory cell order to use as a condition to find constants

    This looks at the first train for each source, issuing a warning if the
    pattern differs between sources.
    """
    patterns = []
    for source in sources:
        try:
            cell_id_data = run[source, 'image.cellId'].drop_empty_trains()
        except SourceNameError:
            continue  # Source missing

        if len(cell_id_data.train_ids) == 0:
            continue  # No data for this module
        cell_ids = cell_id_data[0].ndarray().flatten()
        if not any(np.array_equal(cell_ids, p) for p in patterns):
            patterns.append(cell_ids)

    if len(patterns) > 1:
        warn("Memory cell order varies between detector modules: "
             "; ".join([f"{s[:10]}...{s[-10:]}" for s in patterns]))
    elif not patterns:
        raise ValueError("Couldn't find memory cell order for any modules")

    return patterns[0]


def make_cell_order_condition(use_param, cellid_pattern) -> Optional[str]:
    """Convert the cell ID array to a condition string, or None if not used"""
    if use_param == 'auto':
        # auto -> use cell order if it wraps around (cells not filled monotonically)
        use = len(cellid_pattern) > 2 and (
                np.diff(cellid_pattern.astype(np.int32)) < 0
        ).any()
    else:
        use = (use_param == 'always')

    return (",".join([str(c) for c in cellid_pattern]) + ",") if use else None


def sort_dark_runs_by_gain(raw_folder, runs, ctrl_src):
    """Check gain factors [100, 10, 1] to decide,
    if provided dark runs are in the correct order.
    
    Args:
        raw_folder (Union[str, Path]): The raw data path.
        runs (list): A list of 3 runs.
        ctrl_src (str): The CTRL source to check `RUN/.../femAsicGain`.
    Return:
        (list): Sorted dark runs.
    """
    run_to_gain = dict()
    expected_gain_factors = [100, 10, 1]
    assert len(set(runs)) == 3, f"A list of {len(runs)} runs is provided, three different dark runs are expected." # noqa

    for r in runs:
        run_to_gain[r] = RunDirectory(
            Path(raw_folder) / f"r{r:04d}").get_run_value(
                ctrl_src, "femAsicGain")

    if len(set(run_to_gain.values())) < 3:
        raise ValueError(
            f"Incorrect gain factors for the provided dark runs: {run_to_gain}."
            f" The expected gain factors for these runs are {expected_gain_factors}.")

    sorted_run_to_gain = dict(sorted(
        run_to_gain.items(),
        key=lambda x: expected_gain_factors.index(x[1])))

    sorted_runs = list(sorted_run_to_gain.keys())
    if list(run_to_gain.values()) != expected_gain_factors:
        warning(
            "Dark runs were incorrectly sorted and "
            f"have now been corrected to: {sorted_runs}.")

    return sorted_runs
