# pylint: disable=missing-module-docstring
import sys
from pathlib import Path
from typing import Callable, Dict, List, NamedTuple, Optional, Union
from unittest import mock

import extra_data.tests.make_examples as make_examples
from pytest_subprocess import FakeProcess

from xfel_calibrate import calibrate, settings


class FakeProcessCalibrate(FakeProcess):
    """Class used to create a fake process which will mock calls to processes
    expected to be called by pycalibration
    """

    def __init__(self):
        """Sets up fake process for use with xfel calibrate

        Fakes:
        - Slurm free nodes call
        - Slurm preempt nodes call
        - Slurm sbatch call
        - Git describe calls (for pulling repo version)

        Expected use is something like:

        ```
        @pytest.fixture(scope="class", autouse=True)
        def fake_process_calibrate(self):
            with FakeProcessCalibrate() as fake_process:
                yield fake_process
        ```
        """
        super().__init__()
        #  Fake calls to slurm
        self.register_subprocess(
            ["sbatch", self.any()], stdout=["000000"]
        )

        #  For the version insertion...
        self.register_subprocess(
            ["git", self.any(), "describe", "--tag"], stdout=["0.0.0"]
        )
        # For recording installed packages
        self.register_subprocess(
            [self.any(), '-m', 'pip', 'freeze'], stdout=["h5py==3.3.0"],
        )
        # For getting Python version from user venv
        self.register_subprocess(
            [self.any(), '--version'], stdout=["Python 3.x.y"]
        )

        self.keep_last_process(True)


class MockProposal:
    """Class used with pytest to create proposal directories with placeholder
    files

    Use this to generate a skeleton of the proposal structure for tests which
    only rely on the files being present and not their contents, e.g. balancing
    by sequences done in some notebooks
    """

    def __init__(
        self,
        *,
        tmp_path: Path,
        runs: Union[int, list, slice],
        instrument: str,
        cycle: str,
        proposal: str,
        sequences: int = 2,
    ):
        """Create a MockProposal object, this should be used to create a fixture

        Expected use is something like:

        ```python3
        @pytest.fixture
        def mock_proposal(tmp_path):
            return MockProposal(tmp_path, runs=4)
        ```

        Then `mock_proposal` can be passed as a fixture (or set to autouse) to
        make the object available for individual or class tests

        Args:
            tmp_path (Path): Temporary path, should come from pytest fixture
            runs (Union[int, list, slice]): Defines what run directories to make
            instrument (str, optional): Instrument name, e.g. "AGIPD".
            cycle (str, optional): Cycle, e.g. "202031".
            proposal (str, optional): Proposal number, e.g. "p900113".
            sequences (int, optional): Number of sequence files. Defaults to 2.
        """
        self.tmp_path = tmp_path
        self.instrument = instrument
        self.cycle = cycle
        self.proposal = proposal
        self.sequences = list(
            range(sequences)
        )  # TODO: Implement once it's in extra-data

        self.path = tmp_path / instrument / cycle / proposal
        self.path_raw = self.path / "raw"
        self.path_proc = self.path / "proc"
        self.path_usr = self.path / "usr"
        self.path_scratch = self.path / "scratch"

        self.path_usr.mkdir(parents=True)
        self.path_scratch.mkdir()

        self.runs = self.create_runs(runs)

    def create_runs(self, runs: Union[int, List[int], slice]) -> Dict[int, Path]:
        """Create run directories with skeleton files for the proposal

        Args:
            runs (Union[int, list, slice]): Defines what run directories to make

        Returns:
            [Dict[int, Path]]: Dictionary of the run number and run directory
        """
        if isinstance(runs, int):
            runs = list(range(runs))
        elif isinstance(runs, list):
            if not all(isinstance(r, int) for r in runs):
                raise TypeError("lists of runs must contain only integers")
        elif isinstance(runs, slice):
            if runs.stop:
                runs = list(range(runs.stop)[runs])
            else:
                raise ValueError("runs slice must have a stop value")
        else:
            raise TypeError("runs must be an int, list of integers, or a slice")

        run_directories = {run: self.path_raw / f"r{run:04d}" for run in runs}

        for run_path in run_directories.values():
            run_path.mkdir(parents=True)

            self._create_run(run_path)

        return run_directories

    @property
    def _create_run(self) -> Callable:
        """Return the number of modules for a detector type

        Returns:
            modules [Iterable]: Number of modules for a detector type
        """
        return {
            "AGIPD": make_examples.make_spb_run,
            "DSSC": make_examples.make_scs_run,
            "JGFR": make_examples.make_jungfrau_run,
            "LPD": make_examples.make_fxe_run,
            "None": lambda x: None,
        }[self.instrument]


class CalibrateCall:
    """Class used with pytest to create call `xfel-calibrate` and log its output

    Use this to manage calls to `xfel-calibrate` via pytest fixtures
    """

    def __init__(
        self,
        reports_dir,
        capsys,
        *,
        detector: str,
        cal_type: str,
        command: str = "xfel-calibrate",
        in_folder: Optional[Path] = None,
        out_folder: Optional[Path] = None,
        extra_args: List[str] = None,
    ):
        """Create a CallibrateCall object, this should be used to create a fixture

        Expected use is something like:

        ```python3
        @pytest.fixture(scope="function")
        def calibrate_call(self, mock_proposal: MockProposal, capsys, tmp_path):
            return CalibrateCall(
                tmp_path,
                capsys,
                in_folder=mock_proposal.path_raw,
                out_folder=mock_proposal.path_proc,
                command="xfel-calibrate",
                detector="AGIPD",
                cal_type="CORRECT",
                extra_args=["--run", "0"],
            )
        ```

        Args:
            tmp_path ([type]): Temporary path, should come from pytest fixture
            capsys ([type]): capsys path, should come from pytest fixture
            detector (str): Detector passed to the command, e.g. AGIPD
            cal_type (str): Calibration type passed to the command, e.g. CORRECT
            command (str): Main entrypoint called. Defaults to "xfel-calibrate".
            in_folder (Optional[Path], optional): Path to the input folder, usually
                raw. Defaults to None.
            out_folder (Optional[Path], optional): Path to the output folder, usually
                proc. Defaults to None.
            extra_args (List[str], optional): Additional arguments to pass to the
                command. Defaults to None.
        """
        self.reports_dir = reports_dir
        self.in_folder = in_folder
        self.out_folder = out_folder

        self.args = [
            command, detector, cal_type, '--report-to', str(reports_dir / 'test.pdf')
        ]
        if in_folder:
            self.args.extend(["--in-folder", str(self.in_folder)])
        if out_folder:
            self.args.extend(["--out-folder", str(self.out_folder)])

        if extra_args:
            self.args.extend(extra_args)

        calibrate.run(argv=self.args)

        out, err = capsys.readouterr()

        self.out: str = out
        self.err: str = err

        Paths = NamedTuple(
            "Paths",
            [
                ("notebooks", List[Path]),
                ("run_calibrate", Path),
                ("finalize", Path),
                ("InputParameters", Path),
            ],
        )

        self.paths = Paths(
            notebooks=list(self.reports_dir.glob("**/*/*.ipynb")),
            run_calibrate=list(self.reports_dir.glob("**/*/run_calibrate.sh"))[0],
            finalize=list(self.reports_dir.glob("**/*/finalize.py"))[0],
            InputParameters=list(self.reports_dir.glob("**/*/InputParameters.rst"))[0],
        )
